/*
 * Copyright (c) 1997, 1998, 1999
 *      Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *      $Id: sdla_ng.c,v 1.13 2007/03/27 17:31:05 sangoma Exp $
 */


/*
 ****************************************************************************
 * sdla_ng.c   WANPIPE(tm) Multiprotocol WAN Link Driver.  Main module.
 *
 * Author:      Alex Feldman    <al.feldman@sangoma.com>
 *
 * ===========================================================================
 * Sep 10, 2003 Alex Feldman      Initial version.
 ****************************************************************************
*/
#include <wanpipe_version.h>
#include <wanpipe_includes.h>
#include <sdlapci.h>
#include <wanpipe.h>        /* WANPIPE common user API definitions */
#include <if_wanpipe_common.h>

#define NG_WAN_NODE_TYPE        "wp_old"
#define NG_WAN_COOKIE		917786111
#define NG_WAN_HOOK_DEBUG	"debug"
#define NG_WAN_HOOK_UPPER	"upper"
#define NG_WAN_HOOK_LOWER	"lower"

/* Netgraph control messages */
enum {
        NGM_WAN_GET_PROMISC = 1,      /* get node's promiscuous mode bit */
        NGM_WAN_SET_PROMISC,          /* enable/disable promiscuous mode */
        NGM_WAN_GET_AUTOSRC,          /* get source address override */
        NGM_WAN_SET_AUTOSRC,          /* enable/disable src addr override */
};


static int	wan_ng_rcv_lower(node_p node, struct mbuf *m);
static int	wan_ng_rcv_upper(node_p node, struct mbuf *m);

static ng_constructor_t wan_ng_constructor;
static ng_rcvmsg_t      wan_ng_rcvmsg;
static ng_shutdown_t    wan_ng_shutdown;
static ng_newhook_t     wan_ng_newhook;
static ng_connect_t     wan_ng_connect;
static ng_rcvdata_t     wan_ng_rcvdata;
static ng_disconnect_t  wan_ng_disconnect;

/* List of commands and how to convert arguments to/from ASCII */
static const struct ng_cmdlist wan_ng_cmdlist[] = {
        {
          NG_WAN_COOKIE,
          NGM_WAN_GET_PROMISC,
          "getpromisc",
          NULL,
          &ng_parse_int32_type
        },
        {
          NG_WAN_COOKIE,
          NGM_WAN_SET_PROMISC,
          "setpromisc",
          &ng_parse_int32_type,
          NULL
        },
        {
          NG_WAN_COOKIE,
          NGM_WAN_GET_AUTOSRC,
          "getautosrc",
          NULL,
          &ng_parse_int32_type
        },
        {
          NG_WAN_COOKIE,
          NGM_WAN_SET_AUTOSRC,
          "setautosrc",
          &ng_parse_int32_type,
          NULL
        },
        { 0 }
};

static struct ng_type typestruct = {
        .version =         NG_ABI_VERSION,
        .name =            NG_WAN_NODE_TYPE,
        .constructor =     wan_ng_constructor,
        .rcvmsg =          wan_ng_rcvmsg,
        .shutdown =        wan_ng_shutdown,
        .newhook =         wan_ng_newhook,
        .connect =         wan_ng_connect,
        .rcvdata =         wan_ng_rcvdata,
        .disconnect =      wan_ng_disconnect,
	.cmdlist =         wan_ng_cmdlist,
};

MODULE_DEPEND(wanpipe, netgraph, NG_ABI_VERSION,NG_ABI_VERSION,NG_ABI_VERSION);
NETGRAPH_INIT(wp_old, &typestruct);

#if 0
static int wan_ng_done_init = 0;
#endif

int wan_ng_init_old(sdla_t* card)
{
	int err;
	/*
	 * we have found a node, make sure our 'type' is availabe.
	 */
#if 0  /* taken care of NETGRAPH_INIT */
	if (wan_ng_init_type(card, NULL)){
		return -EINVAL;
	}
#endif
	if ((err = ng_make_node_common(&typestruct, &card->node)) != 0){
		DEBUG_EVENT("%s: Failed to create NETGRAPH node (err=%d)!\n",
                                  card->devname, err);
		return -EINVAL;
	}
	NG_NODE_SET_PRIVATE(card->node, card);
	card->xmitq.ifq_maxlen = IFQ_MAXLEN;
	card->xmitq_hipri.ifq_maxlen = IFQ_MAXLEN;
	mtx_init(&card->xmitq.ifq_mtx, "wp_xmitq", NULL, MTX_DEF);
	mtx_init(&card->xmitq_hipri.ifq_mtx, "wp_xmitq_hipri", NULL, MTX_DEF);
	if ((err = ng_name_node(card->node, card->devname))) {
		DEBUG_EVENT("%s: Failed to assign a netgraph node a name (err = %d)!\n",
                           card->devname, err);
		ng_rmnode_self(card->node);
		NG_NODE_UNREF(card->node);
		card->node = NULL;
		return -EINVAL;
	}
	return 0;
}

int wan_ng_remove_old(sdla_t* card)
{
	if (card->node){
		ng_rmnode_self(card->node);
		NG_NODE_UNREF(card->node);
		card->node = NULL;
	}
	return 0;
}


/***********************************************************************
 * This section contains the methods for the Netgraph interface
 ***********************************************************************/
/*
 * It is not possible or allowable to create a node of this type.
 * If the hardware exists, it will already have created it. */
static int wan_ng_constructor(node_p node)
{
        return (EINVAL);
}

/*
 * give our ok for a hook to be added...
 * If we are not running this should kick the device into life.
 * The hook's private info points to our stash of info about that
 * channel. */
static int wan_ng_newhook(node_p node, hook_p hook, const char *name)
{
        sdla_t* card = NG_NODE_PRIVATE(node);

        /*
         * check if it's our friend the debug hook
         */
        if (strcmp(name, NG_WAN_HOOK_DEBUG) == 0) {
                NG_HOOK_SET_PRIVATE(hook, NULL);
                card->debug_hook = hook;
                return (0);
        }

	if (!strcmp(name, NG_WAN_HOOK_UPPER)) {
		card->upper = hook;
		card->upperhooks++;
	} else if (!strcmp(name, NG_WAN_HOOK_LOWER)) {
		card->lower = hook;
		card->lowerhooks++;
	} else {
		DEBUG_EVENT("%s: invalid hook!\n",
					card->devname);
		return (EINVAL);
	}
        NG_HOOK_SET_PRIVATE(hook, card);
        if (card->wan_up) card->wan_up(card);

        return (0);
}



/*
 * incoming messages.
 * Just respond to the generic TEXT_STATUS message
 */
static  int
wan_ng_rcvmsg(node_p node, item_p item, hook_p lasthook)
{
        sdla_t* card = NG_NODE_PRIVATE(node);
        struct ng_mesg* resp_msg = NULL;
        struct ng_mesg* msg;
	struct wan_dev_le *devle;
	netdevice_t* ifp = NULL;
        u_char want;
        int error = 0;

        NGI_GET_MSG(item, msg);
        switch (msg->header.typecookie) {
        case NGM_GENERIC_COOKIE:
                switch(msg->header.cmd) {
                case NGM_TEXT_STATUS:
                        {
                        char    *arg;
                        int pos = 0;
                        int resplen;

                        resplen = sizeof(struct ng_mesg) + NG_TEXTRESPONSE;
                        NG_MKRESPONSE(resp_msg, msg, resplen, M_NOWAIT);
                        if (resp_msg == NULL) {
                                error = ENOMEM;
                                break;
                        }
                        arg = (resp_msg)->data;

                        /*
                         * Put in the throughput information.
                         */
                        pos = sprintf(arg, "%ld bytes in, %ld bytes out\n",
                                        card->inbytes, card->outbytes);
                        pos += sprintf(arg + pos,
                                "%ld output errors\n",
                                        card->oerrors);
                        pos += sprintf(arg + pos,
                                "ierrors = %ld", card->ierrors);
                        resp_msg->header.arglen = pos+1;
                        }
                        break;
               default:
                        error = EINVAL;
                        break;
                }
                break;
	case NG_WAN_COOKIE:
		DEBUG_TEST("%s: Received WAN cookie type!\n",
						card->devname);
		switch(msg->header.cmd) {
		case NGM_WAN_GET_PROMISC:
			NG_MKRESPONSE(resp_msg, msg, sizeof(u_int32_t), M_NOWAIT);
			if (resp_msg == NULL) {
				error = ENOMEM;
				break;
			}
			*((u_int32_t *)resp_msg->data) = card->promisc;
			break;
		case NGM_WAN_SET_PROMISC:
                        if (msg->header.arglen != sizeof(u_int32_t)) {
                                error = EINVAL;
                                break;
                        }
                        want = !!*((u_int32_t *)msg->data);
                        if (want ^ card->promisc) {
				WAN_LIST_FOREACH(devle, &card->wandev.dev_head, dev_link) {
					ifp = WAN_DEVLE2DEV(devle);
					if (ifp) {
						if ((error = ifpromisc(ifp, want)) != 0)
							continue;
						card->promisc = want;
					}
				}
			}
			break;
		case NGM_WAN_GET_AUTOSRC:
			NG_MKRESPONSE(resp_msg, msg, sizeof(u_int32_t), M_NOWAIT);
			if (resp_msg == NULL) {
                                error = ENOMEM;
                                break;
                        }
                        *((u_int32_t *)resp_msg->data) = card->autoSrcAddr;
			break;
		case NGM_WAN_SET_AUTOSRC:
			if (msg->header.arglen != sizeof(u_int32_t)) {
                                error = EINVAL;
                                break;
                        }
                        card->autoSrcAddr = !!*((u_int32_t *)msg->data);
                        break;
		default:
			error = EINVAL;
			break;
		};
		break;
        default:
                //error = EINVAL;
                break;
        }
        NG_RESPOND_MSG(error, node, item, resp_msg);
        NG_FREE_MSG(msg);
        return (error);
}

/*
 * get data from another node and transmit it to the correct channel
 */
static int wan_ng_rcvdata(hook_p hook, item_p item)
{
        node_p		node = NG_HOOK_NODE(hook);
        sdla_t*		card = NG_NODE_PRIVATE(node);
        struct mbuf*    m;
        int             error = 0;

        NGI_GET_M(item, m);
        NGI_GET_META(item, meta);
        NG_FREE_ITEM(item);

        /*
         * data doesn't come in from just anywhere (e.g control hook)
         */
        if (NG_HOOK_PRIVATE(hook) == NULL) {
                error = ENETDOWN;
                goto bad;
        }
 	if (hook == card->lower) {
		return (wan_ng_rcv_lower(node, m));
	}
	if (hook == card->upper) {
		return (wan_ng_rcv_upper(node, m));
	}

	/* invalid hook */
	DEBUG_TEST("%s: ERROR: received packet NOT belonging to an hook!\n",
					card->devname);
	return (0);

bad:
        /*
         * It was an error case.
         * check if we need to free the mbuf, and then return the error
         */
        NG_FREE_M(m);
	return (error);
}

/*
 * Process data set for tranmission to device.
 */
static int  wan_ng_rcv_lower(node_p node, struct mbuf *m)
{
        sdla_t			*card = NG_NODE_PRIVATE(node);
	netdevice_t		*ifp = NULL;
	wanpipe_common_t        *common;
	wan_spinlock_t          s;
        int			error = 0;

	ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
	if (ifp == NULL){
		error = EINVAL;
                goto bad;
        }
	common = wan_netif_priv(ifp);
        WAN_ASSERT(common == NULL);
        if (common->iface.output == NULL){
                DEBUG_EVENT("%s: Protocol output function is undefined!\n",
                                wan_netif_name(ifp));
                IF_DROP(&ifp->if_snd);
                wan_skb_free(m);
                return -EINVAL;
        }
	wan_bpf_report(ifp, m, 1, WAN_BPF_DIR_OUT);
        wan_spin_lock(&s);
        error = common->iface.output(ifp, m, NULL, NULL);
        wan_spin_unlock(&s);
        if (error == 1){
		DEBUG_EVENT("%s: Error in protocol output function!\n", 
			    wan_netif_name(ifp));
                IF_DROP(&ifp->if_snd);
                wan_skb_free(m);
                error = ENETDOWN;
        }
	return (error);
bad:
        /*
         * It was an error case.
         * check if we need to free the mbuf, and then return the error
         */
        NG_FREE_M(m);
	return (error);
}

/*
 * Process data set for upper layers of IP stack.
 */
static int  wan_ng_rcv_upper(node_p node, struct mbuf *m)
{
	sdla_t*                 card = NG_NODE_PRIVATE(node);
        netdevice_t*            ifp = NULL;
	wanpipe_common_t        *common;
	int                     error = 0;

        ifp = WAN_DEVLE2DEV(WAN_LIST_FIRST(&card->wandev.dev_head));
        if (ifp == NULL){
        	error = EINVAL;
                goto bad;
        }
	common = wan_netif_priv(ifp);
	WAN_ASSERT(common == NULL);
	/* If we are sending it up, remove our M_FLAG */
	wan_skb_clear_mark(m);

	if (common->is_netdev){
		netisr_queue(
			(m->m_flags & M_PROTO1)?NETISR_IPX:NETISR_IP, m); 
	}else{
		WAN_SPPP_INPUT(ifp, m);
	}
	return (0);

bad:
        /*
         * It was an error case.
         * check if we need to free the mbuf, and then return the error
         */
        NG_FREE_M(m);
        return (error);
}

/*
 * do local shutdown processing..
 * this node will refuse to go away, unless the hardware says to..
 * don't unref the node, or remove our name. just clear our links up.
 */
static int wan_ng_shutdown(node_p node)
{
        sdla_t* card = NG_NODE_PRIVATE(node);

        if (card->wan_down) card->wan_down(card);
        return (0);
}

/* already linked */
static int wan_ng_connect(hook_p hook)
{
        /* be really amiable and just say "YUP that's OK by me! " */
        return (0);
}

/*
 * Notify on hook disconnection (destruction)
 *
 * Invalidate the private data associated with this dlci.
 * For this type, removal of the last link resets tries to destroy the node.
 * As the device still exists, the shutdown method will not actually
 * destroy the node, but reset the device and leave it 'fresh' :)
 *
 * The node removal code will remove all references except that owned by the
 * driver.
 */
static int wan_ng_disconnect(hook_p hook)
{
        node_p  node = NG_HOOK_NODE(hook);
        sdla_t* card = NG_NODE_PRIVATE(node);
        int     s;

        /*
         * If it's the data hook, then free resources etc.
         */
	WAN_ASSERT(card == NULL);
	if (hook == card->upper) {
		s = splimp();
		card->upperhooks--;
		splx(s);
	} else if (hook == card->lower) {
		s = splimp();
		card->lowerhooks--;
		splx(s);
	} else {
		/* what? */
	}

        return (0);
}
